import datetime
from typing import Dict, Optional, Tuple

from jatime.converter import (
    ja_hour_to_24_hour,
    ja_num_to_int,
    jp_year_to_ad_year,
    relative_day_into_absolute,
    relative_month_into_absolute,
    relative_year_into_absolute,
)
from jatime.errors import InvalidValueError
from jatime.finder import year_from_month_day_dow, year_month_from_day_dow


class DateTime(object):
    ATTRS = ["year", "month", "day", "dow", "hour", "minute"]
    DOW = "月火水木金土日"

    def __init__(self, base: Optional[datetime.datetime] = None, **kwargs) -> None:
        self.base = datetime.datetime.now() if base is None else base

        self.year = self._given_year(**kwargs)
        self.month = self._given_month(**kwargs)
        self.day = self._given_day(**kwargs)
        self.dow = self._given_dow(**kwargs)
        self.hour = self._given_hour(**kwargs)
        self.minute = self._given_minute(**kwargs)

        self._estimate()

    def _relative_year(self, **kwargs) -> Optional[int]:
        if "relative_year" not in kwargs.keys():
            return None

        year = relative_year_into_absolute(kwargs["relative_year"], self.base.year)
        if year is None:
            return None

        return year

    def _given_year(self, **kwargs) -> Optional[int]:
        year = self._relative_year(**kwargs)
        if year is not None:
            return year

        year_ad = ja_num_to_int(kwargs.get("ad_year", None))

        if "jp_year" not in kwargs.keys():
            return year_ad

        year_jp = jp_year_to_ad_year(kwargs["jp_year"])
        if year_jp is None:
            return year_ad
        elif year_ad is None:
            return year_jp

        if year_jp == year_ad:
            return year_jp
        else:
            raise InvalidValueError(
                f"{kwargs['jp_year']} is not {year_ad}, but {year_jp}."
            )

    def _relative_month(self, **kwargs) -> Optional[Tuple[int, int]]:
        if "relative_month" not in kwargs.keys():
            return None

        year_month = relative_month_into_absolute(
            kwargs["relative_month"], self.base.year, self.base.month
        )
        if year_month is None:
            return None

        return year_month

    def _given_month(self, **kwargs) -> Optional[int]:
        year_month = self._relative_month(**kwargs)
        if year_month is not None:
            self.year = year_month[0]
            return year_month[1]

        month = ja_num_to_int(kwargs.get("month", None))
        if month is None:
            return None
        if month < 1 or 12 < month:
            raise InvalidValueError(f"month must be in 1..12, but {month} was given.")

        return month

    def _relative_day(self, **kwargs) -> Optional[Tuple[int, int, int]]:
        if "relative_day" not in kwargs.keys():
            return None

        year_month_day = relative_day_into_absolute(
            kwargs["relative_day"], self.base.year, self.base.month, self.base.day
        )
        if year_month_day is None:
            return None

        return year_month_day

    def _given_day(self, **kwargs) -> Optional[int]:
        year_month_day = self._relative_day(**kwargs)
        if year_month_day is not None:
            self.year = year_month_day[0]
            self.month = year_month_day[1]
            return year_month_day[2]

        day = ja_num_to_int(kwargs.get("day", None))
        if day is None:
            return None
        if day < 1 or 31 < day:
            raise InvalidValueError(f"day must be in 1..31, but {day} was given.")
        return day

    def _given_dow(self, **kwargs) -> Optional[int]:
        if self.year is not None and self.month is not None and self.day is not None:
            date = datetime.date(self.year, self.month, self.day)
            calculated_dow = date.weekday()
        else:
            date = None
            calculated_dow = None

        if "dow" not in kwargs.keys():
            return calculated_dow

        dow = self.DOW.index(kwargs["dow"])
        if calculated_dow is None or dow == calculated_dow:
            return dow

        raise InvalidValueError(
            f"day of the week on {date} is not {dow}, but {calculated_dow}."
        )

    @staticmethod
    def _given_hour(**kwargs) -> Optional[int]:
        hour = ja_num_to_int(kwargs.get("hour", None))
        if hour is None:
            return None
        if "ampm" in kwargs.keys():
            hour = ja_hour_to_24_hour(kwargs["ampm"], hour)
        if hour < 0 or 29 < hour:
            raise InvalidValueError(f"hour must be in 0..29, but {hour} was given.")
        return hour

    @staticmethod
    def _given_minute(**kwargs) -> Optional[int]:
        minute = ja_num_to_int(kwargs.get("minute", None))
        if minute is None:
            return None
        if minute < 0 or 59 < minute:
            raise InvalidValueError(f"minute must be in 0..59, but {minute} was given.")
        return minute

    def _estimate(self) -> None:
        if (
            self.year is None
            and self.month is not None
            and self.day is not None
            and self.dow is not None
        ):
            self.year = year_from_month_day_dow(
                self.month, self.day, self.dow, self.base.year
            )

        if (
            self.year is None
            and self.month is None
            and self.day is not None
            and self.dow is not None
        ):
            self.year, self.month = year_month_from_day_dow(
                self.day, self.dow, self.base.year, self.base.month
            )

    def to_dict(self) -> Dict:
        return {key: getattr(self, key) for key in self.ATTRS}



__title__ = "jatime"
__description__ = "Time Representation Analyzer for Japanese."
__url__ = "https://github.com/poyo46/jatime"
__version__ = "0.1.0"
__author__ = "poyo46"
__author_email__ = "poyo4rock@gmail.com"
__license__ = "Apache 2.0"
__copyright__ = "Copyright 2020 poyo46"



import datetime
from typing import Generator, Optional, Tuple

from jatime.errors import NotFoundError


def _years_close_to(
    base_year: int, max_diff: Optional[int] = None
) -> Generator[int, None, None]:
    if max_diff is None:
        max_diff = 50

    year = base_year
    for i in range(max_diff * 2):
        if i % 2 == 0:
            year -= i
        else:
            year += i
        yield year


def year_from_month_day_dow(
    month: int, day: int, dow: int, base_year: Optional[int] = None
) -> int:
    """Find the closest year that meets the requirements for a given month,
    day and day of the week.

    Parameters
    ----------
    month : int
        Month.
    day : int
        Day.
    dow : int
        Day of the week as an integer, where Monday is 0 and Sunday is 6.
    base_year : int, optional
        The base year of the search (the default is the year at runtime).

    Returns
    -------
    int
        The the closest year that meets
        ``datetime(year, month, day).weekday() == dow```.

    Raises
    ------
    NotFoundError
        If the year is not found.

    Notes
    -----
    Given a combination of month, day and day of the week that exist, the year is
    always found.

    Examples
    --------
    October 17, 2018 is a Wednesday (dow = 2).

    >>> year_from_month_day_dow(10, 17, 2, 2020)
    2018

    Note that October 17, 2029 is also a Wednesday, but 2018 will be adopted because
    it is closer to 2020.
    """
    if base_year is None:
        base_year = datetime.date.today().year

    for year in _years_close_to(base_year):
        # MEMO: It is always found within +-40 from the base year.
        try:
            date = datetime.date(year, month, day)
            if date.weekday() == dow:
                return year
        except ValueError:
            # February 29th of a non-leap year.
            continue

    raise NotFoundError(
        f"could not find the year: ({month}, {day}, {dow}; {base_year})"
    )


def _years_months_close_to(
    base_year: int, base_month: int, max_diff: Optional[int] = None
) -> Generator[Tuple[int, int], None, None]:
    if max_diff is None:
        max_diff = 50

    month = base_month
    for m in range(max_diff * 2):
        if m % 2 == 0:
            month -= m
        else:
            month += m
        q, r = divmod(month, 12)
        if r != 0:
            next_year = base_year + q
            next_month = r
        else:
            next_year = base_year + q - 1
            next_month = 12
        yield next_year, next_month


def year_month_from_day_dow(
    day: int,
    dow: int,
    base_year: Optional[int] = None,
    base_month: Optional[int] = None,
) -> Tuple[int, int]:
    """Find the closest year and month that meets the requirements for a given day and
    day of the week.

    Parameters
    ----------
    day : int
        Day.
    dow : int
        Day of the week as an integer, where Monday is 0 and Sunday is 6.
    base_year : int, optional
        The base year of the search (the default is the year at runtime).
    base_month : int, optional
        The base month of the search (the default is the month at runtime).

    Returns
    -------
    int
        The the closest year and month that meets
        ``datetime(year, month, day).weekday() == dow```.

    Raises
    ------
    NotFoundError
        If the year or month is not found.

    Notes
    -----
    Given a combination of day and day of the week that exist, the year and month are
    always found.

    Examples
    --------
    February 17, 2021 is a Wednesday (dow = 2).

    >>> year_month_from_day_dow(17, 2, 2020, 12)
    (2021, 2)

    Note that June 17, 2020 is also a Wednesday, but February 2021 will be adopted
    because it is closer to December 2020.
    """
    if base_year is None:
        base_year = datetime.date.today().year
    if base_month is None:
        base_month = datetime.date.today().month

    for year, month in _years_months_close_to(base_year, base_month):
        # MEMO: It is always found within +-20 from the base year and base month.
        try:
            if datetime.date(year, month, day).weekday() == dow:
                return year, month
        except ValueError:
            # February 29th of a non-leap year.
            continue

    raise NotFoundError(
        f"could not find the year and month: ({day}, {dow}; {base_year}, {base_month})"
    )



import itertools
import re
from typing import Generator

# 括弧
_OPEN = r"[(（「『【〔［]"
_CLOSE = r"[)）」』】〕］]"

# 日付表記のセパレータ
_SEP = r"/／"

# 西暦4桁表記
_AD_YEAR = r"(?P<ad_year>[12１２一二][\d〇一二三四五六七八九]{3})"

# 和暦表記
_JP_YEAR = (
    r"(?P<jp_year>(明治|大正|昭和|平成|令和)("
    + "|".join(
        [
            r"\d{1,2}",
            r"[二三四五六][十拾]?[一二三四五六七八九]",
            r"[拾十一][一二三四五六七八九]",
            r"[二三四五六][〇十拾]",
            r"一〇",
            r"[一二三四五六七八九十拾]",
            r"元",
        ]
    )
    + "))"
)

# 相対的な年表現
_RELATIVE_YEAR = r"(?P<relative_year>(一昨年|昨年|去年|今年|再来年|来年))"

# 年表現のパターン一覧
_YEAR = [
    [_JP_YEAR, r"年", _OPEN, _AD_YEAR, r"年?", _CLOSE],
    [_JP_YEAR, _OPEN, _AD_YEAR, _CLOSE, r"年?"],
    [_AD_YEAR, r"年", _OPEN, _JP_YEAR, r"年?", _CLOSE],
    [_AD_YEAR, _OPEN, _JP_YEAR, _CLOSE, r"年?"],
    [_JP_YEAR, r"年"],
    [_AD_YEAR, r"年"],
    [_RELATIVE_YEAR],
]
YEAR = [r"\s*".join(y) for y in _YEAR]

MONTH = [
    r"(?P<month>("
    + "|".join(
        [
            r"[1１][012０１２]",
            r"[十一拾][一二]",
            r"一〇",
            r"[一二三四五六七八九十拾]",
            r"[1-9]",
            r"[１-９]",
        ]
    )
    + r"))\s*月",
    r"(?P<relative_month>(先月|今月|再来月|来月))",
]

DAY = [
    r"(?P<day>("
    + "|".join(
        [
            r"[3３][01０１]",
            r"[12１２]\d",
            r"[1-9１-９]",
            r"三[十拾]?一",
            r"二[十拾]?[一二三四五六七八九]",
            r"[拾十一][一二三四五六七八九]",
            r"[二三][十拾〇]",
            r"一〇",
            r"[一二三四五六七八九十拾]",
        ]
    )
    + r"))\s*日",
    r"(?P<relative_day>(一昨日|昨日|今日|本日|明日|明後日))",
]

# 曜日
DOW = [
    r"\s*".join([_OPEN, r"(?P<dow>[月火水木金土日])(曜日|曜|)", _CLOSE]),
    r"(?P<dow>[月火水木金土日])(曜日|曜)",
]

# 時
_HOUR = (
    r"(?P<ampm>(午前|午後|[aａAＡpｐPＰ]\.?[mｍMＭ]\.?|))"
    + r"\s*"
    + r"(?P<hour>("
    + "|".join(
        [
            r"[012０１２]?\d",
            r"二[十拾]?[一二三四五六七八九]",
            r"[拾十一][一二三四五六七八九]",
            r"[二][〇十]",
            r"一〇",
            r"[零〇一二三四五六七八九十拾]",
        ]
    )
    + r"))"
)
# 分
_MINUTE = (
    r"(?P<minute>("
    + "|".join(
        [
            r"[0-5０-５]?\d",
            r"[二三四五][十拾]?[一二三四五六七八九]",
            r"[拾十一][一二三四五六七八九]",
            r"[二三四五][〇十]",
            r"一〇",
            r"[零〇一二三四五六七八九十拾]",
        ]
    )
    + r"))"
)
_TIME = [
    [_HOUR, "[:：]", _MINUTE],
    [_HOUR, "時", _MINUTE, "分"],
    [_HOUR, "時", r"(?P<minute>半?)"],
]
TIME = [r"\s*".join(t) for t in _TIME]


# 優先順位付けられた日付表現のリスト
_ORDERED_PATTERNS = [
    (YEAR, MONTH, DAY, DOW, TIME),
    (YEAR, MONTH, DAY, TIME),
    (YEAR, MONTH, DAY, DOW),
    (YEAR, MONTH, DAY),
    (YEAR, MONTH),
    (YEAR,),
    (MONTH, DAY, DOW, TIME),
    (MONTH, DAY, TIME),
    (MONTH, DAY, DOW),
    (MONTH, DAY),
    (MONTH,),
    (DAY, DOW, TIME),
    (DAY, TIME),
    (DAY, DOW),
    (DAY,),
    (DOW, TIME),
    (TIME,),
    (DOW,),
]


def datetime_patterns() -> Generator:
    for time_repr in _ORDERED_PATTERNS:
        for pattern_tuple in itertools.product(*time_repr):
            yield re.compile(r"\s*の?\s*".join(pattern_tuple))



from flask import Flask, jsonify, request
from werkzeug.exceptions import BadRequest

from jatime import __copyright__, __description__, __license__, __title__, __version__
from jatime.analyzer import analyze

app = Flask("jatime")
app.config["JSON_AS_ASCII"] = False


@app.route("/")
def root():
    return jsonify(
        {
            "name": __title__,
            "description": __description__,
            "version": __version__,
            "license": __license__,
            "copyright": __copyright__,
        }
    )


@app.route("/analysis")
def get_analysis():
    string = request.args.get("string", None)
    if string is None:
        raise BadRequest("Parameter `string` is required.")
    # TODO: allow base time to be specified.
    return jsonify(analyze(string))


@app.after_request
def edit_header(response):
    response.headers["Access-Control-Allow-Origin"] = "*"
    return response


@app.errorhandler(Exception)
def handle_exception(e):
    return jsonify({"error": e.name, "details": e.description}), e.code



import json

import click

from jatime import __version__
from jatime.analyzer import analyze
from jatime.server import app


@click.group()
@click.version_option(__version__)
def cli():
    pass


@cli.command(name="analyze", help="Analyze Japanese time expressions from the string.")
@click.option("--format-json", is_flag=True, help="Output in json format.")
@click.argument("string")
def _analyze(format_json: bool, string: str) -> None:
    result = analyze(string)
    if format_json:
        click.echo(json.dumps(result, indent=2, ensure_ascii=False))
    else:
        click.echo(result)


@cli.command(help="Start an HTTP server.")
@click.option("--host", default="localhost", show_default=True)
@click.option("--port", type=int, default=1729, show_default=True)
def serve(host: str, port: int):
    app.run(host=host, port=port)



import datetime
import re
from typing import Dict, List, Optional, Union

from jatime.errors import InvalidValueError
from jatime.patterns import datetime_patterns
from jatime.times import DateTime


def _split(string: str, pattern) -> List:
    pieces = []
    i = 0
    for m in re.finditer(pattern, string):
        pieces.append(string[i : m.span()[0]])
        pieces.append(m)
        i = m.span()[1]
    if string[i:] != "":
        pieces.append(string[i:])
    return pieces


def split(string: str, patterns) -> List:
    pieces = [string]
    for pattern in patterns:
        new_pieces = []
        for p in pieces:
            if type(p) == str:
                new_pieces += _split(p, pattern)
            else:
                new_pieces.append(p)
        pieces = new_pieces
    return pieces


def analyze(
    string: str, base: Optional[datetime.datetime] = None
) -> List[Union[str, Dict]]:
    if base is None:
        base = datetime.datetime.now()

    result = []
    pieces = split(string, datetime_patterns())
    for p in pieces:
        if type(p) == str:
            result.append(p)
            continue

        dic = {"string": p.group()}
        try:
            dt = DateTime(base=base, **p.groupdict())
            dic.update(dt.to_dict())
        except InvalidValueError as e:
            dic["error"] = str(e)
        result.append(dic)
        # TODO: update base time (?)
    return result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jatime']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'flask>=1.1.2,<2.0.0']

entry_points = \
{'console_scripts': ['jatime = jatime.cui:cli']}

setup_kwargs = {
    'name': 'jatime',
    'version': '0.1.0',
    'description': 'Time expression analyzer for Japanese.',
    'long_description': '# Jatime: Time Expression Analyzer for Japanese\n\n[![PyPI Version](https://img.shields.io/pypi/v/jatime.svg)](https://pypi.org/pypi/jatime/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/jatime.svg)](https://pypi.org/pypi/jatime/)\n[![License](https://img.shields.io/pypi/l/jatime.svg)](https://pypi.org/pypi/jatime/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![Test Status](https://github.com/poyo46/jatime/workflows/Test/badge.svg)](https://github.com/poyo46/jatime/actions?query=workflow%3ATest) \n\nJatime is a Python library to extract Japanese time expressions from a given text and turning them into objects. \n[Try Jatime now from your browser!](https://poyo46.github.io/jatime/)\n\n## Installing jatime\n\nInstall with `pip` or your favorite PyPI package manager.\n\n```\n$ pip install jatime\n```\n\n## Using jatime\n\nHere\'s how to analyze Japanese text using jatime. \nNote that "dow" means the day of the week.\n\n**From within Python**\n\n```python\nfrom jatime.analyzer import analyze\n\n\nresult = analyze("それは令和２年十月十七日の出来事でした。")\nprint(result)\n```\n\n```\n[\'それは\', {\'string\': \'令和２年十月十七日\', \'year\': 2020, \'month\': 10, \'day\': 17, \'dow\': 5, \'hour\': None, \'minute\': None}, \'の出来事でした。\']\n```\n\n**From the command line**\n\n```console\n$ jatime analyze --format-json それは令和２年十月十七日の出来事でした。\n[\n  "それは",\n  {\n    "string": "令和２年十月十七日",\n    "year": 2020,\n    "month": 10,\n    "day": 17,\n    "dow": 5,\n    "hour": null,\n    "minute": null\n  },\n  "の出来事でした。"\n]\n```\n\n**Over HTTP communication**\n```console\n$ jatime serve\n```\n\n```\n$ curl --get "http://localhost:1729/analysis" --data-urlencode "string=それは令和２年十月十七日の出来事でした。"\n["それは",{"day":17,"dow":5,"hour":null,"minute":null,"month":10,"string":"令和２年十月十七日","year":2020},"の出来事でした。"]\n```\n\n## Features\n* Jatime does not change the given string at all. It just objectifies the extracted time expressions.\n* Jatime supports a variety of time expressions. `22:30`, `午後十時半` and `P.M.10:30` are all interpreted as the same time as we want.\n* Jatime tries to make up for the missing time information whenever possible. That is, jatime calculates the year given the month, day and day of the week.\n* Jatime will tell us why the time information is inconsistent if it is inconsistent.\n',
    'author': 'poyo46',
    'author_email': 'poyo4rock@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://poyo46.github.io/jatime/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import datetime
from typing import Callable, Dict, Optional, Tuple, Union


def ja_num_to_int(num: Union[int, str]) -> Optional[int]:
    """Convert the Japanese numerical expression to int.

    Notes
    -----
    * "半" is specially converted to 30.
    * This function supports expressions with 4 or fewer digits, not including
      "百" and "千".

    Parameters
    ----------
    num : int or str
        Japanese numerical expression.

    Returns
    -------
    int or None
        The int value corresponding to ``num``, or None if `it cannot be converted.

    Examples
    --------
    >>> [ja_num_to_int(n) for n in (1, "零", "六七", "八十九", "二〇二〇", "半", "あ")]
    [1, 0, 67, 89, 2020, 30, None]
    """
    if type(num) == int:
        return num

    num = str(num)
    if len(num) == 0:
        return None

    specials = {"零": 0, "元": 1, "十": 10, "拾": 10, "半": 30}
    if num in specials.keys():
        return specials[num]

    num = num[:-1] + num[-1].replace("十", "0").replace("拾", "0")
    num = num[0] + num[1:].replace("十", "").replace("拾", "")
    num = num.translate(str.maketrans("〇一二三四五六七八九十拾", "012345678911"))

    try:
        return int(num)
    except ValueError:
        return None


def jp_year_to_ad_year(jp_year: str) -> Optional[int]:
    """Convert the Japanese year to the Western calendar.

    Notes
    -----
    If a year beyond the final year is given, return None.

    Parameters
    ----------
    jp_year : str
        Japanese year.

    Returns
    -------
    int or None
        The Western calendar corresponding to ``jp_year``, or None if it cannot be
        converted.

    Examples
    --------
    >>> [jp_year_to_ad_year(y) for y in ("昭和二十", "平成4", "令和元", "あ")]
    [1945, 1992, 2019, None]

    Note that the 昭和 era is up to '64.

    >>> assert jp_year_to_ad_year("昭和65") is None
    """
    era = jp_year[:2]
    year = jp_year[2:]
    data = {
        # 年号: (最終年, 最終年に対応する西暦年)
        "明治": (45, 1912),
        "大正": (15, 1926),
        "昭和": (64, 1989),
        "平成": (31, 2019),
        "令和": (50, 2068),
    }
    if era not in data.keys():
        return None
    year = ja_num_to_int(year)
    if year is None:
        return None
    diff = data[era][0] - year
    if diff < 0:
        return None
    return data[era][1] - diff


def _relative_to_absolute(data: Dict[str, int]) -> Callable[[str, int], Optional[int]]:
    def converter(relative_expression: str, base: int) -> Optional[int]:
        assert type(base) == int
        if relative_expression not in data.keys():
            return None
        return base + data[relative_expression]

    return converter


def relative_year_into_absolute(relative_year: str, base_year: int) -> Optional[int]:
    """Convert a relative expression of the year into an absolute one.

    Parameters
    ----------
    relative_year : str
        Relative year expression.
    base_year : int
        Base year.

    Returns
    -------
    int or None
        The western calendar corresponding to ``relative_year`` based on ``base_year``,
        or None if it cannot be converted.

    Examples
    --------
    >>> relative_year_into_absolute("来年", 2020)
    2021

    >>> relative_year_into_absolute("一昨年", 2020)
    2018
    """
    data = {
        "一昨年": -2,
        "昨年": -1,
        "去年": -1,
        "今年": 0,
        "来年": 1,
        "再来年": 2,
    }
    return _relative_to_absolute(data)(relative_year, base_year)


def relative_month_into_absolute(
    relative_month: str, base_year: int, base_month: int
) -> Optional[Tuple[int, int]]:
    """Convert a relative expression of the month into an absolute one.

    Parameters
    ----------
    relative_month : str
        Relative month expression.
    base_year : int
        Base year.
    base_month : int
        Base month.

    Returns
    -------
    tuple of int
        The absolute year and month corresponding to ``relative_month`` based on
        ``base_year`` and ``base_month`, or None if it cannot be converted.

    Examples
    --------
    >>> relative_month_into_absolute("先月", 2021, 1)
    (2020, 12)

    >>> relative_month_into_absolute("今月", 2021, 1)
    (2021, 1)

    >>> relative_month_into_absolute("再来月", 2020, 12)
    (2021, 2)
    """
    assert type(base_year) == int
    data = {
        "先々月": -2,
        "先月": -1,
        "今月": 0,
        "来月": 1,
        "再来月": 2,
    }
    month = _relative_to_absolute(data)(relative_month, base_month)
    if month is None:
        return None
    if month < 1:
        return base_year - 1, month + 12
    elif month <= 12:
        return base_year, month
    else:
        return base_year + 1, month - 12


def relative_day_into_absolute(
    relative_day: str, base_year: int, base_month: int, base_day: int
) -> Optional[Tuple[int, int, int]]:
    """Convert a relative expression of the day into an absolute one.

    Parameters
    ----------
    relative_day : str
        Relative day expression.
    base_year : int
        Base year.
    base_month : int
        Base month.
    base_day : int
        Base day.

    Returns
    -------
    tuple of int
        The absolute year, month and day corresponding to ``relative_day`` based on
        ``base_year``, ``base_month`` and ``base_day``, or None if it cannot be
        converted.

    Examples
    --------
    >>> relative_day_into_absolute("明日", 2020, 10, 17)
    (2020, 10, 18)

    >>> relative_day_into_absolute("昨日", 2020, 3, 1)
    (2020, 2, 29)

    >>> relative_day_into_absolute("昨日", 2021, 3, 1)
    (2021, 2, 28)
    """
    data = {"一昨日": -2, "昨日": -1, "今日": 0, "本日": 0, "明日": 1, "明後日": 2}
    if relative_day not in data.keys():
        return None
    base = datetime.date(base_year, base_month, base_day)
    date = base + datetime.timedelta(days=data[relative_day])
    return date.year, date.month, date.day


def ja_hour_to_24_hour(ampm: str, hour: int) -> int:
    """Unify the expression of hour with 24-hour clock.

    Parameters
    ----------
    ampm : str
        A string representing AM or PM.
    hour : int
        Hour.

    Returns
    -------
        24-hour.

    Examples
    --------
    >>> ja_hour_to_24_hour("午前", 9)
    9

    >>> ja_hour_to_24_hour("午後", 9)
    21

    >>> ja_hour_to_24_hour("AM", 10)
    10

    >>> ja_hour_to_24_hour("PM", 10)
    22
    """
    if "午後" in ampm or ampm.startswith(("p", "ｐ", "P", "Ｐ")):
        hour += 12
    return hour
